#include <AIV/Helpers.mqh>

class CSets
{
  protected : 

    int _syncTime;
  
  protected : 

    CHelpers * _helpers;

  public : 

    CSets(void)
    {
      _helpers = new CHelpers();
    }

    ~CSets(void)
    {
      delete(_helpers);
    }
    
    /**
     * Initialize data set
     */
    void setupSet(SSets &series, const int cols)
    {
      series.mBet.mPlus = 0;
      series.mBet.mMinus = 0;
      series.mBet.mCount = 0;

      series.mSymbol.mMean = 0;
      series.mSymbol.mUnit = 0;
      series.mSymbol.mOrder = 0;
      
      ArrayResize(series.mAsk, cols);
      ArrayResize(series.mBid, cols);
      ArrayResize(series.mLow, cols);
      ArrayResize(series.mHigh, cols);
      ArrayResize(series.mOpen, cols);
      ArrayResize(series.mClose, cols);
      ArrayResize(series.mPoint, cols);
      ArrayResize(series.mPrice, cols);
      ArrayResize(series.mSpread, cols);
      ArrayResize(series.mVolume, cols);
      ArrayResize(series.mTime, cols);
      
      ZeroMemory(series.mAsk);
      ZeroMemory(series.mBid);
      ZeroMemory(series.mLow);
      ZeroMemory(series.mHigh);
      ZeroMemory(series.mOpen);
      ZeroMemory(series.mClose);
      ZeroMemory(series.mPoint);
      ZeroMemory(series.mPrice);
      ZeroMemory(series.mSpread);
      ZeroMemory(series.mVolume);
      ZeroMemory(series.mTime);
    }
    
    /**
     * Copy matrices
     */
    void setupSets(SSets &series[], const int rows, const int cols)
    {
      ArrayResize(series, rows);
      
      for (int k = 0; k < rows; k++)
      {
          setupSet(series[k], cols);
      }
    }
    
    /**
     * Copy array of structures
     */
    void copySets(SSets &source[], SSets &destination[])
    {
      int N = ArraySize(source);
      
      for (int k = 0; k < N; k++)
      {
        destination[k].mSymbol.mName = source[k].mSymbol.mName;
        destination[k].mSymbol.mMean = source[k].mSymbol.mMean;
        destination[k].mSymbol.mUnit = source[k].mSymbol.mUnit;
        destination[k].mSymbol.mOrder = source[k].mSymbol.mOrder;
        
        ArrayCopy(destination[k].mAsk, source[k].mAsk);
        ArrayCopy(destination[k].mBid, source[k].mBid);
        ArrayCopy(destination[k].mLow, source[k].mLow);
        ArrayCopy(destination[k].mHigh, source[k].mHigh);
        ArrayCopy(destination[k].mOpen, source[k].mOpen);
        ArrayCopy(destination[k].mClose, source[k].mClose);
        ArrayCopy(destination[k].mPoint, source[k].mPoint);
        ArrayCopy(destination[k].mPrice, source[k].mPrice);
        ArrayCopy(destination[k].mSpread, source[k].mSpread);
        ArrayCopy(destination[k].mVolume, source[k].mVolume);
        ArrayCopy(destination[k].mTime, source[k].mTime);
      }
    }

    /**
     * Get number of minimum available bars in history
     */
    int getSetsBars(SSets& series[], const ENUM_TIMEFRAMES period, const int limit = 0)
    {
      int order = ArraySize(series);
      int bars = TerminalInfoInteger(TERMINAL_MAXBARS);

      for (int k = 0; k < order; k++)
      {
        bars = MathMin(bars, _helpers.getBars(series[k].mSymbol.mName, period, limit));

        if (bars < 1 || SeriesInfoInteger(series[k].mSymbol.mName, period, SERIES_SYNCHRONIZED) == 0)
        {
          return 0;
        }
      }

      return bars;
    }

    /**
     * List combinations of provided currencies
     */
    int getCombinations(SSets& series[], const string names, const string end = "")
    {
      string syms[];
      int index = 0;
      int count = StringSplit(names, ',', syms);

      for (int k = 0; k < count; k++)
      {
        for (int n = 0; n < count; n++)
        {
          string s1 = syms[k] + syms[n] + end;
          string s2 = syms[n] + syms[k] + end;
          bool c1 = SymbolSelect(s1, true);
          bool c2 = SymbolSelect(s2, true);
          bool c3 = seriesByName(series, s1) < 0;
          bool c4 = seriesByName(series, s2) < 0;
          bool c5 = SymbolInfoInteger(s1, SYMBOL_TRADE_MODE) == SYMBOL_TRADE_MODE_FULL;
          bool c6 = SymbolInfoInteger(s2, SYMBOL_TRADE_MODE) == SYMBOL_TRADE_MODE_FULL;

          if ((c1 && c3) || (c2 && c4))
          {
            ArrayResize(series, index + 1);
            series[index].mSymbol.mName = (c1 ? s1 : s2);
            index++;
          }
        }
      }

      return index;
    }
    
    /**
     * Split string of symbols by separator
     */
    int getPairs(SSets& series[], const string names, const bool validate = false)
    {
      string syms[];
      int index = 0;
      int count = StringSplit(names, ',', syms);

      for (int k = 0; k < count; k++)
      {
        string inverse = StringSubstr(syms[k], 3, 3) + StringSubstr(syms[k], 0, 3) + StringSubstr(syms[k], 6);
        bool A = SymbolSelect(syms[k], true) && SymbolInfoInteger(syms[k], SYMBOL_TRADE_MODE) == SYMBOL_TRADE_MODE_FULL;
        bool B = SymbolSelect(inverse, true) && SymbolInfoInteger(inverse, SYMBOL_TRADE_MODE) == SYMBOL_TRADE_MODE_FULL;

        if (A || validate == false)
        {
          ArrayResize(series, index + 1); 
          series[index].mSymbol.mName = syms[k];
          series[index].mSymbol.mOrder = 0;
          index++;
        }
        else if (B || validate == false)
        {
          ArrayResize(series, index + 1); 
          series[index].mSymbol.mName = inverse;
          series[index].mSymbol.mOrder = 1;
          index++;
        }
      }
      
      return index;
    }
    
    /**
     * Format list of deals ordered by profit made
     */
    string getBets(SSets &series[], SBet &results[])
    {
      string bets = "";
      int order = ArraySize(series);
      
      if (order > 0)
      {
        int size = ArraySize(series[0].mPoint);
        ArrayResize(results, size);
    
        for (int n = 0; n < size; n++)
        {
          int count = 0;
          int plus = 0;
      
          for (int k = 0; k < order; k++)
          {
            if (series[k].mPoint[n] < 2)
            {
              count++;
              if (series[k].mPoint[n] == 1) plus += int(series[k].mPoint[n]);
            }
          }

          results[n].mPlus = plus;
          results[n].mCount = count;
          results[n].mMinus = count - plus;
  
          bets += 
            IntegerToString(n) + " : " + 
            IntegerToString(count) + " : " + 
            IntegerToString(plus) + " : " + 
            IntegerToString(count - plus) + (n < size - 1 ? "\n" : "");
        }
      }
      
      return bets;
    }
    
    /**
     * Get last time bar in a series
     */
    datetime getSetsTime(SSets &series[], const ENUM_TIMEFRAMES period)
    {
      datetime startTime = 0;
      int count = ArraySize(series);
  
      for (int k = 0; k < count; k++)
      {
        startTime = MathMax(startTime, _helpers.getLastSymbolTime(series[k].mSymbol.mName));
      }
  
      return startTime;
    }

    /** 
     * Get quotes without synchronization
     */
    int getQuotes(SSets& series, const ENUM_TIMEFRAMES period, const int position, const int count)
    {
      MqlRates rates[];

      int bars = Bars(series.mSymbol.mName, period);
      int items = CopyRates(series.mSymbol.mName, period, position, count, rates);

      if (items < 1)
      {
        Print(
          "Synchronization : " + series.mSymbol.mName + ", " + 
          "Position : " + IntegerToString(position) + ", " + 
          "Depth : " + IntegerToString(count));

        return 0;
      }

      int index = 0;

      setupSet(series, count);

      for (int n = items - 1; n > -1; n--)
      {
        series.mLow[index] = rates[n].low;
        series.mHigh[index] = rates[n].high;
        series.mOpen[index] = rates[n].open;
        series.mClose[index] = rates[n].close;
        series.mPoint[index] = series.mClose[index];
        series.mSpread[index] = rates[n].spread;
        series.mVolume[index] = rates[n].real_volume ? rates[n].real_volume : rates[n].tick_volume;
        series.mTime[index] = rates[n].time;

        index++;
      }

      return items;
    }

    /**
     * Get list of quotes sunchronized by date
     */
    int getQuotesMatches(
      SSets &series, 
      const ENUM_TIMEFRAMES period, 
      const int position = 0, 
      const int count = 0) 
    {
      MqlRates rate;
      MqlRates rates[];

      setupSet(series, count);

      datetime start = _helpers.positionToTime(position, period);

      int seconds = PeriodSeconds(period);
      int secondsInMonth = PeriodSeconds(PERIOD_MN1);
      int precision = _helpers.getPrecision(series.mSymbol.mName);
      int items = CopyRates(series.mSymbol.mName, period, start, count, rates);

      if (items < 1 || SeriesInfoInteger(series.mSymbol.mName, period, SERIES_SYNCHRONIZED) == 0)
      {
        Print("### Error : No price for " + series.mSymbol.mName + " from " + TimeToString(position));
        return 0;
      }

      int index = items - 1;
      datetime date = rates[index].time;

      rate = rates[index];

      for (int k = 0; k < count; k++) 
      {
        series.mLow[k] = rate.low;
        series.mHigh[k] = rate.high;
        series.mOpen[k] = rate.open;
        series.mClose[k] = rate.close;
        series.mPoint[k] = series.mClose[k];
        series.mTime[k] = rate.time;
        series.mVolume[k] = rate.real_volume ? rate.real_volume : rate.tick_volume;

        date -= seconds;

        if (date == rates[MathMax(index - 1, 0)].time || seconds >= secondsInMonth) 
        {
          index = MathMax(index - 1, 0);
          rate = rates[index];
        }
      }

      return items;
    }    

    /**
     * Synchronize datasets by time
     */
    int getEqualSets(
      SSets &series[], 
      const ENUM_TIMEFRAMES period, 
      const int position, 
      const int count = 1,
      const int sync = 0)
    {
      int order = ArraySize(series);
      int min = TerminalInfoInteger(TERMINAL_MAXBARS);
      int bars = MathMin(getSetsBars(series, period, count), count);

      if (bars < 1)
      {
        return 0;
      }

      for (int k = 0; k < order; k++)
      {
        int counted = sync == 0 ? 
          getQuotes(series[k], period, position, count) : 
          getQuotesMatches(series[k], period, position, count);

        if (counted < 1)
        {
          min = MathMin(min, counted);
        }
      }

      if (min < 1)
      {
        return 0;
      }

      return bars;
    }
    
    /**
     * Search in array of structures by name
     */
    int seriesByName(SSets &series[], const string name)
    {
      int order = ArraySize(series);

      for (int k = 0; k < order; k++)
      {
        if (series[k].mSymbol.mName == name)
        {
          return k;
        }
      }
      
      return -1;
    }
    
    /**
     * Get abstract prices, as is
     */
    void getPrices(SSets& source[], SSets& destination[])
    {
      int order = ArraySize(source);
      int size = ArraySize(source[0].mPrice);

      setupSets(destination, order, size);

      for (int k = 0; k < order; k++)
      {
        destination[k].mSymbol.mName = source[k].mSymbol.mName;

        for (int n = 0; n < size; n++)
        {
          destination[k].mPoint[n] = source[k].mPoint[n];
        }
      }
    }
    
    /**
     * Get normalized abstract prices, as logarithms
     */
    void getPriceLogs(SSets& source[], SSets& destination[])
    {
      int order = ArraySize(source);
      int size = ArraySize(source[0].mPrice);

      setupSets(destination, order, size);

      for (int k = 0; k < order; k++)
      {
        destination[k].mSymbol.mName = source[k].mSymbol.mName;

        for (int n = 0; n < size; n++)
        {
          destination[k].mPoint[n] = MathLog(source[k].mPoint[n]);
        }
      }
    }

    /**
     * Get normalized prices, as percentage change from point zero
     */
    void getPricePercents(SSets& source[], SSets& destination[])
    {
      int order = ArraySize(source);
      int size = ArraySize(source[0].mPrice);

      setupSets(destination, order, size);

      for (int k = 0; k < order; k++)
      {
        destination[k].mSymbol.mName = source[k].mSymbol.mName;

        if (source[k].mSymbol.mUnit == 0)
        {
          source[k].mSymbol.mUnit = source[k].mPoint[0];
        }

        for (int n = 0; n < size; n++)
        {
          destination[k].mPoint[n] = (source[k].mPoint[n] - source[k].mSymbol.mUnit) / source[k].mSymbol.mUnit;
        }
      }
    }

    /**
     * Get prices in deposit's currency
     */
    void getPriceCurrencies(SSets& source[], SSets& destination[])
    {
      int order = ArraySize(source);
      int size = ArraySize(source[0].mPrice);

      setupSets(destination, order, size);

      for (int k = 0; k < order; k++)
      {
        destination[k].mSymbol.mName = source[k].mSymbol.mName;

        if (source[k].mSymbol.mUnit == 0)
        {
          source[k].mSymbol.mUnit = source[k].mPoint[0];
        }

        for (int n = 0; n < size; n++)
        {
          destination[k].mPoint[n] = (source[k].mPoint[n] - source[k].mSymbol.mUnit) * _helpers.getTickCost(source[k].mSymbol.mName);
        }
      }
    }
            
    /**
     * Switch series to scaled chart from -1 to 1
     */
    void getPriceLimits(SSets& source[], SSets& destination[], double start = 0, double end = 1)
    {
      int order = ArraySize(source);
      int size = ArraySize(source[0].mPrice);

      setupSets(destination, order, size);

      for (int k = 0; k < order; k++)
      {
        double min = 1000000;
        double max = -1000000;

        destination[k].mSymbol.mMean = 0;

        for (int n = 0; n < size; n++)
        {
          destination[k].mSymbol.mMean += source[k].mPoint[n];
          min = MathMin(source[k].mPoint[n], min);
          max = MathMax(source[k].mPoint[n], max);
        }

        destination[k].mSymbol.mMean /= size;
        destination[k].mSymbol.mName = source[k].mSymbol.mName;

        for (int n = 0; n < size; n++)
        {
          destination[k].mPoint[n] = 0;

          if (max - min != 0)
          {
            destination[k].mPoint[n] = start + (source[k].mPoint[n] - min) * (end - start) / (max - min);
          }
        }
      }
    }

    /**
     * Get prices aligned by volatility
     */
    void getPriceVolatilities(SSets& source[], SSets& destination[])
    {
      int order = ArraySize(source);
      int size = ArraySize(source[0].mPrice);
      int min = getMinIndex(source);

      setupSets(destination, order, size);

      for (int k = 0; k < order; k++)
      {
        destination[k].mSymbol.mName = source[k].mSymbol.mName;
      
        if (source[k].mSymbol.mUnit == 0)
        {
          source[k].mSymbol.mUnit = source[k].mPoint[size - 1];
        }
      }

      for (int k = 0; k < order; k++)
      {
        double factor = source[min].mSymbol.mUnit / source[k].mSymbol.mUnit;

        for (int n = 0; n < size; n++)
        {
          destination[k].mPoint[n] = (source[k].mPoint[n] - source[k].mSymbol.mUnit) * _helpers.getTickCost(source[k].mSymbol.mName) * factor;
        }
      }
    }
    
    /**
     * Get series with lowest volatility
     */
    int getMinIndex(SSets &series[])
    {
      int minIndex = 0;
      int order = ArraySize(series);

      double min = 1000000;
      double max = -1000000;

      for (int k = 0; k < order; k++)
      {
        double currentMin = series[k].mPoint[0];
        double currentMax = series[k].mPoint[0];

        if (MathAbs(min) > MathAbs(currentMin) && MathAbs(max) > MathAbs(currentMax))
        {
          min = currentMin;
          max = currentMax;
          minIndex = k;
        }
      }

      return minIndex;
    }

    /**
     * Get synthetics neutral to the provided symbol
     */
    int getCustomSymbols( 
      const string names, 
      SName& combinations[],
      const string nameBase, 
      const string nameQuote)
    {
      string symbols[];
      SName combosBase[];
      SName combosQuote[];

      int codes[];
      int size = 0;
      int sizeBase = 0;
      int sizeQuote = 0;
      int count = StringSplit(names, ',', symbols);

      ArrayResize(codes, count);
      ArrayResize(combosBase, (int) MathPow(2, count));
      ArrayResize(combosQuote, (int) MathPow(2, count));
      ArrayResize(combinations, (int) MathPow(2, 15));

      ZeroMemory(codes);
      ZeroMemory(combosBase);
      ZeroMemory(combosQuote);
      ZeroMemory(combinations);

      while (_helpers.getChain(codes, count))
      {
        string name = NULL;

        for (int k = 0; k < count; k++)
        {
          if (codes[k])
          {
            name = _helpers.isEmpty(name) ? symbols[k] : name + "," + symbols[k];
          }
        }

        if (_helpers.isEmpty(nameBase) || StringFind(name, nameBase) >= 0) 
        {
          combosBase[sizeBase].mData = name;
          sizeBase++;
        }

        if (_helpers.isEmpty(nameQuote) || StringFind(name, nameQuote) >= 0) 
        {
          combosQuote[sizeQuote].mData = name;
          sizeQuote++;
        }
      }

      for (int k = 0; k < sizeBase; k++)
      {
        for (int n = 0; n < sizeQuote; n++)
        {
          string namesBase[];
          string namesQuote[];

          int countBase = StringSplit(combosBase[k].mData, ',', namesBase);
          int countQuote = StringSplit(combosQuote[n].mData, ',', namesQuote);

          for (int kk = 0; kk < countBase; kk++)
          {
            for (int nn = 0; nn < countQuote; nn++)
            {
              if (namesBase[kk] == namesQuote[nn])
              {
                namesBase[kk] = NULL;
                namesQuote[nn] = NULL;
              }
            }
          }

          string operationBase = "";

          for (int kk = 0; kk < countBase; kk++)
          {
            operationBase += namesBase[kk];
          }

          string operationQuote = "";

          for (int kk = 0; kk < countQuote; kk++)
          {
            operationQuote += namesQuote[kk];
          }

          if (operationBase == nameBase && operationQuote == nameQuote && countBase == countQuote)
          {
            string duplicates = "#";

            countBase = StringSplit(combosBase[k].mData, ',', namesBase);
            countQuote = StringSplit(combosQuote[n].mData, ',', namesQuote);

            for (int kk = 0; kk < countBase; kk++)
            {
              for (int nn = 0; nn < countQuote; nn++)
              {
                string name = NULL;

                for (int ii = 0; ii < countQuote; ii++)
                {
                  string symbol = namesQuote[nn];
                  namesQuote[nn] = namesQuote[ii];
                  namesQuote[ii] = symbol;
                }

                bool valid = true;
                int countData = MathMax(countBase, countQuote);

                for (int ii = 0; ii < countData; ii++)
                {
                  string dataBase = countBase > ii ? namesBase[ii] : "";
                  string dataQuote = countQuote > ii ? namesQuote[ii] : "";

                  string symbolBase = _helpers.isEmpty(dataBase) ? dataQuote : dataBase + ":" + dataQuote;
                  string symbolQuote = _helpers.isEmpty(dataQuote) ? dataBase : dataQuote + ":" + dataBase;

                  bool symbolsDuplicate = StringFind(name, symbolQuote) >= 0;
                  bool symbolsEqual = dataBase == dataQuote;
                  bool symbolsSame = dataBase == nameBase && dataQuote == nameQuote;

                  if (symbolsDuplicate || symbolsEqual || symbolsSame)
                  {
                    valid = false;
                    break;
                  }

                  name += _helpers.isEmpty(name) ? symbolBase : "," + symbolBase;
                }

                if (valid && _helpers.isEmpty(name) == false && StringFind(duplicates, name) < 0)
                {
                  combinations[size++].mData = name;
                  duplicates += name + "#";
                }
              }
            }
          }
        }
      }

      ArrayResize(combinations, size);

      for (int k = 0; k < size; k++)
      {
        string combos[];
        int countSymbols = StringSplit(combinations[k].mData, ',', combos);
        ArrayResize(combinations[k].mSeries, countSymbols);

        for (int n = 0; n < countSymbols; n++)
        {
          string parts[];
          int countParts = StringSplit(combos[n], ':', parts);
          string symbolBase = parts[0] + parts[1];
          string symbolQuote = parts[1] + parts[0];

          if (SymbolSelect(symbolBase, true))
          {
            combinations[k].mSeries[n].mSymbol.mName = symbolBase;
            combinations[k].mSeries[n].mSymbol.mOrder = 1;
          }
          else
          {
            combinations[k].mSeries[n].mSymbol.mName = symbolQuote;
            combinations[k].mSeries[n].mSymbol.mOrder = -1;
          }
        }
      }

      return size;
    }
};